import math

import glm
import numpy as np
from OpenGL.GL import *

from math_helper import MathHelper


class LineDrawer:

    def __init__(self, program, projViewMatrix):
        # Save the program and projView matrix
        self.program = program
        self.projViewMatrix = projViewMatrix

        # Activate the current shader program to access shader variables
        glUseProgram(self.program)

        # Access the uniform variables in the shaders
        self.aPositionLocation = glGetAttribLocation(self.program, "aPosition")
        self.uColorLocation = glGetUniformLocation(self.program, "uColor")
        self.uMvpMatrixLocation = glGetUniformLocation(self.program, "uMvpMatrix")

        # Create a buffer in the video card's RAM
        self.vertPosBuffer = glGenBuffers(1)
        self.initVertexBuffers()

    def initVertexBuffers(self):
        # Set the vertices of the square
        vertPositions = np.array([
            -0.5, -0.5, 0.5,
            -0.5, 0.5, 0.5,
            -0.5, 0.5, -0.5,
            -0.5, 0.5, 0.5,
            0.5, 0.5, 0.5,
            0.5, 0.5, -0.5,
            0.5, 0.5, 0.5,
            0.5, -0.5, 0.5,
            0.5, -0.5, -0.5,
            0.5, -0.5, 0.5,
            -0.5, -0.5, 0.5,
            -0.5, -0.5, -0.5,
            -0.5, -0.5, -0.5,
            -0.5, 0.5, -0.5,
            0.5, 0.5, -0.5,
            0.5, -0.5, 0.5,
            0.5, 0.5, 0.5,
            -0.5, 0.5, 0.5,
            -0.5, -0.5, -0.5,
            -0.5, -0.5, 0.5,
            -0.5, 0.5, -0.5,
            -0.5, 0.5, -0.5,
            -0.5, 0.5, 0.5,
            0.5, 0.5, -0.5,
            0.5, 0.5, -0.5,
            0.5, 0.5, 0.5,
            0.5, -0.5, -0.5,
            0.5, -0.5, -0.5,
            0.5, -0.5, 0.5,
            -0.5, -0.5, -0.5,
            0.5, -0.5, -0.5,
            -0.5, -0.5, -0.5,
            0.5, 0.5, -0.5,
            -0.5, -0.5, 0.5,
            0.5, -0.5, 0.5,
            -0.5, 0.5, 0.5
        ], dtype=np.float32)
        # Bind to the created buffer
        glBindBuffer(GL_ARRAY_BUFFER, self.vertPosBuffer)
        # Copy vertex array to buffer
        glBufferData(GL_ARRAY_BUFFER, len(vertPositions) * 4,
            vertPositions, GL_STATIC_DRAW)

    # This method should be called if the window
    # size changes or the camera position changes.
    def setProjViewMatrix(self, projViewMatrix):
        self.projViewMatrix = projViewMatrix

    def bind(self):
        # Activate Shader Program Object
        glUseProgram(self.program)

        # Bind to the buffer
        glBindBuffer(GL_ARRAY_BUFFER, self.vertPosBuffer)

        # Set up buffer
        glVertexAttribPointer(self.aPositionLocation, 3, GL_FLOAT, GL_FALSE,
            0, ctypes.c_void_p(0))
        glEnableVertexAttribArray(self.aPositionLocation)

    def draw(self, start, end, color, thickness):
        # Find the center of the segment
        v = end - start
        centerPosition = start + v / 2

        # Find the length of the segment
        length = glm.length(v)

        # Normalize the segment vector
        norm = glm.normalize(v);

        # Calculate the angle of a segment
        rotation = MathHelper.rotationTo(glm.vec3(1, 0, 0), norm);

        # Create a model matrix, that is, a matrix combining the
        # translation matrix, rotation matrix, and the scale matrix
        # Create a translation matrix
        modelMatrix = glm.translate(glm.mat4(1), centerPosition)
        # Create a rotation matrix
        rotationMatrix = glm.mat4_cast(rotation)
        modelMatrix = modelMatrix * rotationMatrix
        # Create a scale matrix
        modelMatrix = glm.scale(modelMatrix, glm.vec3(length, thickness, thickness))

        # Combine projView matrix and model matrix into one MVP matrix
        mvpMatrix = self.projViewMatrix * modelMatrix

        self.bind()

        # Send MVP matrix to the vertex shader
        glUniformMatrix4fv(self.uMvpMatrixLocation, 1, GL_FALSE,
            glm.value_ptr(mvpMatrix))

        # Send color value to fragment shader
        glUniform3fv(self.uColorLocation, 1, glm.value_ptr(color))

        # Call a draw command that will cause the vertex shader
        # to be called 4 times - once for each vertex of the square
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 36)
